@@ -1,5 +1,6 @@
-######################################################################
-# editing methods
+"""
+Editing methods.
+"""
 
 from __future__ import nested_scopes
 import re, string, time
@@ -7,13 +8,17 @@ from string import split,join,find,lower,rfind,atoi,strip
 from urllib import quote, unquote
 from types import *
 from email.Message import Message
+from copy import deepcopy
 
 import ZODB # need this for pychecker
 from AccessControl import getSecurityManager, ClassSecurityInfo
 from App.Common import rfc1123_date
 from DateTime import DateTime
 from Globals import InitializeClass
-from OFS.content_types import guess_content_type
+try:
+    from zope.app.content_types import guess_content_type
+except ImportError:
+    from OFS.content_types import guess_content_type
 from OFS.DTMLDocument import DTMLDocument
 from OFS.ObjectManager import BadRequestException
 import OFS.Image
@@ -22,20 +27,20 @@ from pagetypes import PAGETYPES
 from Defaults import DISABLE_JAVASCRIPT, LARGE_FILE_SIZE, LEAVE_PLACEHOLDER
 import Permissions
 from Regexps import javascriptexpr, htmlheaderexpr, htmlfooterexpr
-from Utils import BLATHER, parseHeadersBody
-from UI import onlyBodyFrom
-from I18nSupport import _
+from Utils import get_transaction, BLATHER, parseHeadersBody
+from Views import onlyBodyFrom
+from I18n import _
 
 
-class EditingSupport:
+class PageEditingSupport:
     security = ClassSecurityInfo()
     security.declareObjectProtected('View')
     def checkPermission(self, permission, object):
         return getSecurityManager().checkPermission(permission,object)
 
     security.declareProtected(Permissions.Add, 'create') 
-    def create(self,page,text=None,type=None,title='',REQUEST=None,log='',
-               sendmail=1, parents=None, subtopics=None):
+    def create(self,page=None,text='',type=None,title='',REQUEST=None,log='',
+               sendmail=1, parents=None, subtopics=None, pagename=None):
         """
         Create a new wiki page, with optional extras.
 
@@ -44,7 +49,9 @@ class EditingSupport:
         We assume the page name comes url-quoted. If it's not a url-safe
         name, we will create the page with a similar url-safe id, which we
         assume won't match any existing page (or zwiki would have linked
-        instead of offering to create). Other features:
+        instead of offering to create). Also it allows the alternate pagename
+        argument, to support the page management form (XXX temporary).
+        Other features:
 
         - can upload a file at the same time.  
 
@@ -56,7 +63,11 @@ class EditingSupport:
         - redirects to the new page if appropriate
 
         """
-        name = unquote(page)
+        if not self.checkSufficientId(REQUEST):
+            return self.denied(
+                _("Sorry, this wiki doesn't allow anonymous edits. Please configure a username in options first."))
+
+        name = unquote(page or pagename)
         id = self.canonicalIdFrom(name)
 
         # here goes.. sequence is delicate here
@@ -70,6 +81,7 @@ class EditingSupport:
         newid = self.folder()._setObject(id,p)
         p = getattr(self.folder(),newid)
 
+        p.checkForSpam(text)
         p.setCreator(REQUEST)
         p.setLastEditor(REQUEST)
         p.setLastLog(log)
@@ -80,7 +92,7 @@ class EditingSupport:
 
         # choose the specified type, the default type or whatever we're allowed 
         p.setPageType(type or self.defaultPageType())
-        p.setText(text or '',REQUEST)
+        p.setText(text,REQUEST)
         p.handleFileUpload(REQUEST)
         p.handleSubtopicsProperty(subtopics,REQUEST)
         if p.autoSubscriptionEnabled(): p.subscribeThisUser(REQUEST)
@@ -90,7 +102,7 @@ class EditingSupport:
         # everything gets handled properly.  We must first commit though,
         # to get p.cb_isMoveable(). Renaming will also get us indexed.
         # We never leave a placeholder, always update backlinks.
-        if title and title != p.title_or_id():
+        if title and title != p.pageName():
             get_transaction().note('rename during creation')
             get_transaction().commit()
             p.handleRename(title,0,1,REQUEST,log)
@@ -104,7 +116,7 @@ class EditingSupport:
                 p.read(),
                 REQUEST=REQUEST,
                 subjectSuffix='',
-                subject='(%s) %s' % (_('new'),log),
+                subject=log,
                 message_id=self.messageIdFromTime(p.creationTime())
                 )
         # and move on
@@ -135,6 +147,10 @@ class EditingSupport:
         zope subject attribute.  note and use_heading are not used and
         kept only for backwards compatibility.
         """
+        if not self.checkSufficientId(REQUEST):
+            return self.denied(
+                _("Sorry, this wiki doesn't allow anonymous edits. Please configure a username in options first."))
+
         if self.isDavLocked(): return self.davLockDialog()
 
         # gather various bits and pieces
@@ -173,13 +189,18 @@ class EditingSupport:
         # types.
         # add to source, in standard rfc2822 format:
         t = str(m)
-        if self.usingPurpleNumbers(): t = self.addPurpleNumbersTo(t,self)
+        self.checkForSpam(t)
         t = '\n\n' + t
         self.raw += t
-        # add to html cache, rendered:
+        # add to prerendered html:
+        # apply single-message prerendering XXX
         t = self.pageType().preRenderMessage(self,m)
-        if self.messageCount()==1:t=self.pageType().discussionSeparator(self)+t
+        # if it's the first, add appropriate discussion separator
+        if self.messageCount()==1:
+            t=self.pageType().discussionSeparator(self) + t
+        # apply page's standard prerender to the lot XXX
         t = self.pageType().preRender(self,t)
+        # and append to the page's prerendered html
         self.setPreRendered(self.preRendered()+t)
         self.cookDtmlIfNeeded()
 
@@ -256,8 +277,18 @@ class EditingSupport:
         elif self.pageWithNameOrId(page):
             p = self.pageWithNameOrId(page) # changing another page
         else:
-            return self.create(page,text,type,title,REQUEST,log,
+            return self.create(page,
+                               text or '', # string expected here
+                               type,
+                               title,
+                               REQUEST,
+                               log,
                                subtopics=subtopics) # creating a page
+
+        if not self.checkSufficientId(REQUEST):
+            return self.denied(
+                _("Sorry, this wiki doesn't allow anonymous edits. Please configure a username in options first."))
+
         if check_conflict:
             if self.checkEditConflict(timeStamp, REQUEST):
                 return self.editConflictDialog()
@@ -268,7 +299,7 @@ class EditingSupport:
         # handlers checks permissions and does the necessary.
         if p.handleDeleteMe(text,REQUEST,log): return
         p.handleEditPageType(type,REQUEST,log)
-        p.handleEditText(text,REQUEST,subjectSuffix,log)
+        if text != None: p.handleEditText(text,REQUEST,subjectSuffix,log)
         p.handleSubtopicsProperty(subtopics,REQUEST)
         p.handleFileUpload(REQUEST,log)
         p.handleRename(title,leaveplaceholder,updatebacklinks,REQUEST,log)
@@ -335,29 +366,34 @@ class EditingSupport:
 
 
     def handleEditText(self,text,REQUEST=None, subjectSuffix='', log=''):
-        # is the new text valid and different ?
-        if (text is not None and
-            self.cleanupText(text) != self.read()):
+        old = self.read()
+        new = self.cleanupText(text)
+        # is the new text different ?
+        if new != old:
             # do we have permission ?
             if (not
                 (self.checkPermission(Permissions.Edit, self) or
                  (self.checkPermission(Permissions.Append, self)
-                  and find(self.cleanupText(text),self.read()) == 0))):
+                  and find(new,old) == 0))):
                 raise 'Unauthorized', (
                     _('You are not authorized to edit this ZWiki Page.'))
 
+            # does this edit look like spam ?
+            # Tries to count the links added by this edit. Not perfect -
+            # existing links on a line that you tweak will be counted.
+            # Not sure what happens if you replace existing links.
+            self.checkForSpam(self.addedText(old, new))
+                
             # change it
-            oldtext = self.read()
             self.setText(text,REQUEST)
             self.setLastEditor(REQUEST)
             self.setLastLog(log)
 
-            # if mailout policy is all edits, do it here
-            if getattr(self,'mailout_policy','') == 'edits':
-                self.sendMailToSubscribers(
-                    self.textDiff(a=oldtext,b=self.read()),
-                    REQUEST=REQUEST,
-                    subject=log)
+            # send mail if appropriate
+            self.sendMailToEditSubscribers(
+                self.textDiff(a=old,b=self.read()),
+                REQUEST=REQUEST,
+                subject='(edit) %s' % log)
 
     def handleDeleteMe(self,text,REQUEST=None,log=''):
         if not text or not re.match('(?m)^DeleteMe', text):
@@ -408,10 +444,6 @@ class EditingSupport:
 
         XXX no it won't ? also, need a flag to disable reparenting
         """
-        if not (self.checkPermission(Permissions.Delete, self) and
-                self.requestHasSomeId(REQUEST)):
-            raise 'Unauthorized', (
-                _('You are not authorized to delete this ZWiki Page.'))
         oldname, oldid = self.pageName(), self.getId()
         # update parents attributes to avoid orphans
         self.moveMyChildrenTo(self.primaryParentName())
@@ -419,20 +451,19 @@ class EditingSupport:
         if pagename and string.strip(pagename):
             self.replaceLinksThroughoutWiki(oldname,pagename,REQUEST)
         # get parent url while we still can
-        parenturl = self.primaryParentUrl()
+        redirecturl = self.upUrl()
         # unindex (and remove from outline) and move to the recycle bin folder
         self.recycle(REQUEST)
         # notify subscribers if appropriate
-        if getattr(self,'mailout_policy','') == 'edits':
-            self.sendMailToSubscribers(
-                'This page was deleted.\n',
-                REQUEST=REQUEST,
-                subjectSuffix='',
-                subject='(deleted)')
-        if REQUEST: REQUEST.RESPONSE.redirect(parenturl)
+        self.sendMailToEditSubscribers(
+            'This page was deleted.\n',
+            REQUEST=REQUEST,
+            subjectSuffix='',
+            subject='(deleted)')
+        if REQUEST: REQUEST.RESPONSE.redirect(redirecturl)
 
     def ensureRecycleBin(self):
-        if not hasattr(self.folder(),'recycle_bin'):
+        if not hasattr(self.folder().aq_base,'recycle_bin'):
             self.folder().manage_addFolder('recycle_bin', 'deleted wiki pages')
 
     def recycle(self, REQUEST=None):
@@ -470,17 +501,16 @@ class EditingSupport:
         - notify subscribers
         - if called with the existing name, ensures that id conforms to
         canonicalId(title).
+
+        XXX the sendmail argument doesn't stop mailouts from updating backlinks
         """
         # anything to do ?
         oldname, oldid = self.pageName(), self.getId()
+        # newlines would cause glitches later..
+        pagename = re.sub(r'[\r\n]','',pagename)
         newname, newid = pagename, self.canonicalIdFrom(pagename)
         if not newname or (newname == oldname and newid == oldid): return 
 
-        # require a username as well as rename permission for renaming
-        if not self.requestHasSomeId(REQUEST):
-            raise 'Unauthorized', (
-                _('You are not authorized to rename this ZWiki Page.'))
-
         BLATHER('renaming %s (%s) to %s (%s)...'%(oldname,oldid,newname,newid))
 
         # an old page just might have an empty title attribute - set it to
@@ -498,9 +528,10 @@ class EditingSupport:
 
         # update wiki outline
         # changeIdPreservingCreator->manage_renameObject->_delObject/_setObject
-        # has the effect of losing our place in the hierarchy, take a
-        # snapshot so we can fix it up later
-        savedparentmap = self.wikiOutline().parentmap().copy()
+        # -> manage_after* has the effect of losing our place in the
+        # hierarchy, take a snapshot so we can fix it up later
+        savedparentmap = deepcopy(self.wikiOutline().parentmap())
+        savedchildmap = deepcopy(self.wikiOutline().childmap())
 
         # has the page id changed ?
         if newid != oldid:
@@ -518,6 +549,7 @@ class EditingSupport:
         # update wiki outline, using the copy we saved earlier
         # nb that outline may not have been up to date, but replace will forgive
         self.wikiOutline().setParentmap(savedparentmap)
+        self.wikiOutline().setChildmap(savedchildmap)
         self.wikiOutline().replace(oldname,newname)
 
         # do this after the above so it will have correct parent
@@ -533,9 +565,8 @@ class EditingSupport:
             except BadRequestException: pass
 
         # notify subscribers if appropriate
-        if (getattr(self,'mailout_policy','') == 'edits' and
-            sendmail and newname != oldname):
-            self.sendMailToSubscribers(
+        if sendmail and newname != oldname:
+            self.sendMailToEditSubscribers(
                 'This page was renamed from %s to %s.\n'%(oldname,newname),
                 REQUEST=REQUEST,
                 subjectSuffix='',
@@ -601,8 +632,7 @@ class EditingSupport:
                      % (re.escape(oldlink), self.canonicalIdFrom(oldlink))
         newpat = (self.isWikiName(newlink) and newlink) or '[%s]' % newlink
         self.edit(text=re.sub(oldpat, newpat, self.read()),
-                  REQUEST=REQUEST,
-                  log='links updated after rename')
+                  REQUEST=REQUEST)
 
     def handleFileUpload(self,REQUEST,log=''):
         # is there a file upload ?
@@ -650,7 +680,7 @@ class EditingSupport:
                 # failed to create - give up (what about an error)
                 pass
 
-    def _createFileOrImage(self,file,title='',REQUEST=None,parent=None):
+    def _createFileOrImage(self,file,title='',REQUEST=None):
         # based on WikiForNow which was based on
         # OFS/Image.py:File:manage_addFile
         """
@@ -658,34 +688,48 @@ class EditingSupport:
         suffix. Returns a tuple containing the new id, content type &
         size, or (None,None,None).
         """
+        # macro to check folder contents without acquiring
+        folderHas = lambda folder,id: hasattr(folder.aq_base,id)
+
         # set id & title from filename
         title=str(title)
         id, title = OFS.Image.cookId('', title, file)
         if not id:
             return None, None, None
 
-        # find out where to store files - in an 'uploads'
-        # subfolder if defined, otherwise in the wiki folder
-        if (hasattr(self.folder().aq_base,'uploads') and
-            self.folder().uploads.isPrincipiaFolderish):
-            folder = self.folder().uploads
-        else:
-            folder = parent or self.folder() # see create()
+        # find out where to store files - in an 'uploads' subfolder if
+        # present, otherwise the wiki folder
+        folder = self.folder()
+        if (folderHas(folder,'uploads') and
+            folder.uploads.isPrincipiaFolderish):
+            folder = folder.uploads
 
-        if hasattr(folder,id) and folder[id].meta_type in ('File','Image'):
-            pass
-        else:
-            # First, we create the file or image object without data
+        # unless it already exists, create the file or image object
+        # use the CMF/Plone types if appropriate
+        # it will be renamed if there is an id collision with some other
+        # kind of object
+        if not (folderHas(folder,id) and
+                folder[id].meta_type in ('File','Image')):#,
+                                         #'Portal File','Portal Image')):
             if guess_content_type(file.filename)[0][0:5] == 'image':
-                folder._setObject(id, OFS.Image.Image(id,title,''))
+                if self.inCMF():
+                    #XXX how ?
+                    id = folder._setObject(id, OFS.Image.Image(id,title,'')) 
+                else:
+                    id = folder._setObject(id, OFS.Image.Image(id,title,''))
             else:
-                folder._setObject(id, OFS.Image.File(id,title,''))
+                if self.inCMF():
+                    #XXX how ?
+                    id = folder._setObject(id, OFS.Image.File(id,title,''))
+                else:
+                    id = folder._setObject(id, OFS.Image.File(id,title,''))
 
         # Now we "upload" the data.  By doing this in two steps, we
-        # can use a database trick to make the upload more efficient.
-        folder._getOb(id).manage_upload(file)
+        # can use a database trick to make the upload more efficient. (?)
+        ob = folder._getOb(id)
+        ob.manage_upload(file)
 
-        return id, folder._getOb(id).content_type, folder._getOb(id).getSize()
+        return (id, ob.content_type, ob.getSize())
 
     def _addFileLink(self, file_id, content_type, size, REQUEST):
         """
@@ -749,8 +793,6 @@ class EditingSupport:
         Also inserts purple number NIDs if appropriate.
         """
         self.raw = self.cleanupText(text)
-        if self.usingPurpleNumbers(): 
-            self.raw = self.addPurpleNumbersTo(self.raw,self)
         self.preRender(clear_cache=1)
         # re-cook DTML's cached parse data if necessary
         # will prevent edit if DTML can't parse.. hopefully no auth trouble
@@ -761,16 +803,45 @@ class EditingSupport:
         #    #get_transaction().commit()
         #    DTMLDocument.__call__(self,self,REQUEST,REQUEST.RESPONSE)
 
+    def checkForSpam(self, t):
+        """
+        Check for signs of spam in some text, and raise an error if found.
+
+        Also looks at the current user's info in REQUEST.
+        """
+        REQUEST = getattr(self,'REQUEST',None)
+        username = self.usernameFrom(REQUEST,ip_address=0)
+        ip = getattr(REQUEST,'REMOTE_ADDR','')
+        page = self.pageName()
+        def raiseSpamError(reason, verbose_reason):
+            BLATHER(('blocked edit from %s (%s) on %s (%s)\n%s\n') % \
+                    (ip, username, page, reason, t))
+            raise _("There was a problem: %s. Please contact the site administrator for help." % \
+                    (verbose_reason))
+            
+        # banned link pattern ?
+        for pat in getattr(self.folder(),'banned_links',[]):
+            pat = strip(pat)
+            if pat and re.search(pat,t):
+                raiseSpamError(_("banned_links match"),
+                               _("your edit contained a banned link pattern"))
+
+        # anonymous edit with too many urls ?
+        prop = 'max_anonymous_links'
+        # we'll handle either an int or string property
+        if (not self.requestHasUsername(REQUEST) and
+            hasattr(self.folder(), prop)):
+            try: max = int(getattr(self.folder(), prop))
+            except ValueError: max = None
+            if max is not None:
+                if len(re.findall(r'https?://',t)) > max:
+                    raiseSpamError(_("exceeded max_anonymous_links"),
+                                   _("adding of external links by anonymous authors is restricted"))
+
     def cleanupText(self, t):
         """
         Do some cleanup of incoming text, also block spam links.
         """
-        # block spammer links - for now, just give an unexplained site error
-        for pat in getattr(self.folder(),'banned_links',[]):
-            pat = strip(pat)
-            if pat and re.search(pat,t):
-                raise "There was a problem with your edit, please contact the site administrator."
-        
         # strip any browser-appended ^M's
         t = re.sub('\r\n', '\n', t)
 
@@ -817,6 +888,15 @@ class EditingSupport:
             self.last_editor = ''
         self.last_edit_time = DateTime(time.time()).ISO()
 
+    def hasCreatorInfo(self):
+        """
+        True if this page already has creator attributes.
+        """
+        return (hasattr(self,'creator') and
+                hasattr(self,'creation_time') and
+                hasattr(self,'creator_ip'))
+                
+
     def setCreator(self, REQUEST=None):
         """
         record my creator, creator_ip & creation_time
@@ -957,5 +1037,5 @@ class EditingSupport:
         """This wiki's default page type."""
         return self.allowedPageTypes()[0]
     
-InitializeClass(EditingSupport)
+InitializeClass(PageEditingSupport)
 
