@@ -1,7 +1,7 @@
 # zwiki page hierarchy functionality
 # based on original code by Ken Manheimer
 #
-# The OutlineSupport mixin manages the wiki's page hierarchy, making use
+# The PageOutlineSupport mixin manages the wiki's page hierarchy, making use
 # of the more generic Outline class.  It is broken into several smaller
 # mixins to help keep things organized.
 #
@@ -35,13 +35,13 @@
 # -keep it synced during reparent
 # -keep it synced during ZMI operations
 # -check mutators for persistence compatibility ?
-# -refactor OutlineSupport
+# -refactor PageOutlineSupport
 # try dropping parents property/index/metadata
 # upgrade process
 # code cleanup/renaming
 # simplify Outline implementation
 # remove Outline/nesting split ?
-# refactor/simplify OutlineRenderingMixin
+# refactor/simplify OutlineRendering
 
 from __future__ import nested_scopes
 import string, re
@@ -52,16 +52,17 @@ from urllib import quote, unquote
 from AccessControl import ClassSecurityInfo
 import Acquisition
 from App.Common import absattr
-from Globals import InitializeClass
+from Globals import InitializeClass, REPLACEABLE
 import Persistence
+from OFS.SimpleItem import SimpleItem
 
 import Permissions
 from Utils import flatten, BLATHER
-from Defaults import PAGE_METATYPE, SHOW_CURRENT_PAGE_IN_CONTENTS
+from Defaults import PAGE_METATYPE
 from Regexps import bracketedexpr
 import Outline
 
-from I18nSupport import _
+from I18n import _
 
 def deepappend(nesting, page):
     """
@@ -74,8 +75,7 @@ def deepappend(nesting, page):
         else: nesting[-1] = [nesting[-1],page]
     return nesting
 
-class PersistentOutline(
-    Outline.Outline, Acquisition.Implicit, Persistence.Persistent):
+class PersistentOutline(Outline.Outline, SimpleItem):
     """
     I am a persistent version of Outline.
 
@@ -85,15 +85,28 @@ class PersistentOutline(
     else: pchildren = []
     if name not in pchildren: pchildren.append(name)
     childmap[parent] = pchildren
+    or:
+    self._p_changed = 1...
     """
-    pass
+    meta_type = 'Zwiki Outline Cache'
+    # always create with this id:
+    id = 'outline'
+    # these objects get auto-generated by ZWikiPage's manage_afterAdd
+    # causing problems for the ATCT migrator in plone 2.1rc<=3. To avoid
+    # it, allow them to be overwritten without complaint.  Maybe drop this
+    # later though, the outline cache can contain valuable ordering
+    # information ?
+    __replaceable__ = REPLACEABLE
 
-class ParentsPropertyMixin:
+InitializeClass(PersistentOutline)
+
+
+class ParentsProperty:
     """
     I provide the parents property, the old way to store page hierarchy.
 
     For now we keep this property up to date, in sync with the wiki
-    outline, but we'll try removing it completely at some point.
+    outline, but we'll try removing it completely at some point. (?)
 
     Although it's a simple and relatively robust design, it doesn't scale
     so well: generating the hierarchy structure from the parents
@@ -174,15 +187,102 @@ class ParentsPropertyMixin:
             self.setParents(cleanedupparents)
             self.index_object() #XXX only need to update parents index & metadata
 
-InitializeClass(ParentsPropertyMixin) 
+InitializeClass(ParentsProperty) 
 
-class OutlineManagerMixin:
+
+class ShowSubtopicsProperty:
+    """
+    I determine when to display subtopics on a page.
+    """
+    security = ClassSecurityInfo()
+
+    def subtopicsEnabled(self,**kw):
+        """
+        Decide in a complicated way if this page should display it's subtopics.
+
+        First, if the folder has a show_subtopics property (can acquire)
+        and it's false, we will never show subtopics.  Otherwise look for
+        a show_subtopics property
+        - in REQUEST
+        - on the current page
+        - on our primary ancestor pages, all the way to the top,
+        and return the first one we find. Otherwise return true.
+        """
+        prop = 'show_subtopics'
+        if getattr(self.folder(),prop,1):
+            if kw.has_key(prop):
+                return kw[prop] and 1
+            elif hasattr(self,'REQUEST') and hasattr(self.REQUEST,prop):
+                return getattr(self.REQUEST,prop) and 1
+            elif hasattr(self.aq_base,prop):
+                return getattr(self,prop) and 1
+            elif self.primaryParent():
+                # poor caching
+                try: return self.primaryParent().subtopicsEnabled() 
+                except:
+                    # experimental: support all-brains
+                    try: return self.primaryParent().getObject().subtopicsEnabled() 
+                    except: # XXX still run into errors here, investigate
+                        BLATHER('DEBUG: error in subtopicsEnabled for %s, primaryParent is: %s'\
+                             % (self.id(),`self.primaryParent()`))
+                        return not (getattr(getattr(self,'REQUEST',None),
+                                            'zwiki_displaymode',
+                                            None) == 'minimal')
+            else:
+                #return not (getattr(getattr(self,'REQUEST',None),
+                #                    'zwiki_displaymode',
+                #                    None) == 'minimal')
+                return 1
+        else:
+            return 0
+
+    def subtopicsPropertyStatus(self):
+        """
+        Get the status of the show_subtopics property on this page.
+
+        no property:    -1 ("default")
+        true property:   1 ("always")
+        false property:  0 ("never")
+        """
+        if not hasattr(self.aq_base,'show_subtopics'): return -1
+        else: return self.show_subtopics and 1
+
+    def setSubtopicsPropertyStatus(self,status,REQUEST=None):
+        """
+        Set, clear or remove this page's show_subtopics property.
+
+        Same values as getSubtopicsStatus.
+        """
+        props = map(lambda x:x['id'], self._properties)
+        if status == -1:
+            if 'show_subtopics' in props:
+                self.manage_delProperties(ids=['show_subtopics'],
+                                          REQUEST=REQUEST)
+        elif status:
+            if not 'show_subtopics' in props:
+                self.manage_addProperty('show_subtopics',1,'boolean',
+                                        REQUEST=REQUEST)
+            else:
+                self.manage_changeProperties(show_subtopics=1,
+                                             REQUEST=REQUEST)
+        else:
+            if not 'show_subtopics' in props:
+                self.manage_addProperty('show_subtopics',0,'boolean',
+                                        REQUEST=REQUEST)
+            else:
+                self.manage_changeProperties(show_subtopics=0,
+                                             REQUEST=REQUEST)
+
+InitializeClass(ShowSubtopicsProperty)
+
+
+class OutlineManager:
     """
     I manage and query a cached outline object for the wiki.
 
-    When first accessed I generate it based on the parents properties and
-    cache it as a folder attribute.  This is the new way to represent page
-    hierarchy.
+    When first accessed I create a PersistentOutline based on the parents
+    properties and store it in the wiki folder.  This can be queried more
+    efficiently.
 
     old:
     KM: We could make this a persistent object and minimize recomputes.
@@ -197,33 +297,67 @@ class OutlineManagerMixin:
 
     # mutators
 
+    def wikiOutline(self):
+        """
+        Get the outline cache which holds details of the wiki's page hierarchy.
+
+        We'll generate it if needed, ie if it's missing or if it's one of
+        the older types which get lost during a folder rename or when
+        moving pages to a new folder.
+
+        This is called by any method which requires the outline cache, and
+        also by upgrade() on each page view, to ensure that renaming or
+        moving an old wiki just works without requiring the user to
+        manually updateWikiOutline.
+        """
+        if (not hasattr(self.folder().aq_base,'outline')
+            or not self.folder().outline):
+            self.updateWikiOutline()
+        return self.folder().outline
+
     security.declareProtected(Permissions.View, 'updateWikiOutline')
     def updateWikiOutline(self):
         """
         Regenerate the wiki folder's cached outline object.
-        """
-        BLATHER('saving outline data for wiki',self.folder().getId())
-        self.folder().outline = self.wikiOutlineFromParents()
-        REQUEST = getattr(self,'REQUEST',None)
-        if REQUEST: REQUEST.RESPONSE.redirect(self.contentsUrl())
 
-    # easier alias
-    updatecontents = updateWikiOutline
-        
-    def wikiOutlineFromParents(self):
-        """
-        Generate an outline object from the pages' parents properties.
+        The wiki's outline object (a PersistentOutline) is a
+        representation of the page hierarchy, containing the same
+        information as in the pages' parents properties but in a form
+        easier to query. This method either generates a new one from the
+        parents properties, or updates an old one trying to preserve the
+        order of subtopics. Complications.
 
-        We check and correct any invalid parents as we go.  This touches
-        all page objects, which is probably ok as this is not done
-        frequently.
+        This checks and corrects any invalid parents information.  It also
+        loads all page objects from the ZODB, which is probably ok as this
+        is not done frequently.
         """
+        BLATHER('regenerating outline data for wiki',self.folder().getId())
+        oldchildmap = {}
+        # backwards compatibility
+        # there have been three kinds of outline cache: folder attribute,
+        # non-SimpleItem-based PersistentOutline object, and
+        # SimpleItem-based PersistentOutline object
+        # a pre-0.39 outline is just an attribute, delete (but save childmap)
+        if (hasattr(self.folder().aq_base,'outline') and
+            not 'outline' in self.folder().objectIds()):
+            oldchildmap = self.folder().outline.childmap()
+            del self.folder().outline
+        # if there's no outline object, make one
+        if not hasattr(self.folder().aq_base,'outline'):
+            self.folder()._setObject('outline', PersistentOutline())
+            self.folder().outline.setChildmap(oldchildmap)
+        # regenerate the parentmap
         parentmap = {}
         for p in self.pageObjects():
             p.ensureValidParents() # poor caching
             parentmap[p.pageName()] = p.getParents()
-        return PersistentOutline(parentmap)
+        self.folder().outline.setParentmap(parentmap)
+        # update the childmap (without losing subtopics order) and nesting
+        self.folder().outline.update()
 
+    # easier-to-type alias ? XXX remove ? updateoutline ?
+    updatecontents = updateWikiOutline
+        
     security.declareProtected(Permissions.Reparent, 'reparent')
     def reparent(self, parents=[], REQUEST=None, pagename=None):
         """
@@ -237,6 +371,7 @@ class OutlineManagerMixin:
         Page names may be ids, or fuzzy, even partial. Any which do not
         resolve to an existing page or are duplicates will be ignored.
         """
+        oldparents = self.getParents()
         # clean the arguments carefully to avoid parenting anomalies
         # page mgmt form must use pagename field:
         if pagename:
@@ -263,21 +398,31 @@ class OutlineManagerMixin:
         self.wikiOutline().reparent(self.pageName(),uniqueparents)
         self.index_object()
 
+        # send mail if appropriate
+        self.sendMailToEditSubscribers(
+            '%s was reparented from %s to %s.' % (
+              self.pageName(), oldparents, uniqueparents),
+            REQUEST=REQUEST,
+            subject='(reparented)')
+
         if REQUEST is not None: REQUEST.RESPONSE.redirect(REQUEST['URL1'])
 
+    security.declareProtected(Permissions.Reparent, 'reorder')
+    def reorder(self, child, REQUEST=None):
+        """
+        Move child one place to the left among this page's children.
+
+        Ordering is only stored in the wiki outline cache, and will be
+        lost if the outline is completely regenerated.
+
+        When invoked from the web, redirects to /backlinks.
+        """
+        if not child in self.childrenAsList(): return
+        self.wikiOutline().reorder(self.pageName(),child)
+        if REQUEST is not None: REQUEST.RESPONSE.redirect(REQUEST['URL1']+'/backlinks')
+
     # queries
 
-    def wikiOutline(self):
-        """
-        Get the outline object representing the wiki's page hierarchy.
-
-        We'll generate it if needed.
-        """
-        if (not hasattr(self.folder().aq_base,'outline')
-            or not self.folder().outline):
-            self.updateWikiOutline()
-        return self.folder().outline
-
     def primaryParentName(self):
         """
         Get the name of this page's primary (alphabetically first) parent.
@@ -298,9 +443,13 @@ class OutlineManagerMixin:
         """
         p = self.primaryParent()
         if p: return p.pageUrl()
-        # going up from a top level page leads to..
-        #else: return self.wikiUrl()    # the front page
-        else: return self.contentsUrl() # the wiki contents
+        else: return None
+
+    def upUrl(self):
+        """
+        Get the URL of whatever is "above" this page.
+        """
+        return self.primaryParentUrl() or self.defaultPageUrl()
 
     security.declareProtected(Permissions.View, 'firstPage')
     def firstPage(self):
@@ -419,9 +568,10 @@ class OutlineManagerMixin:
         return map(lambda x:absattr(self.pageWithNameOrId(x).id),
                    self.offspringAsList())
 
-    # queries returning nestings (lists-of-lists) - a legacy
+    # queries returning nestings (lists-of-lists) - a low-level
     # representation of all or part of the wiki outline, which can be
-    # processed by the render methods.
+    # processed by certain render methods. See also nestingAsRenderList's
+    # docstring.
 
     def ancestorsNesting(self):
         """
@@ -447,100 +597,16 @@ class OutlineManagerMixin:
         """
         return self.wikiOutline().children(self.pageName())
 
-    def offspringNesting(self):
+    def offspringNesting(self,depth=None):
         """
         Return a nesting representing this page's descendants.
         """
-        return self.wikiOutline().offspring([self.pageName()])
+        return self.wikiOutline().offspring([self.pageName()],depth=depth)
 
-InitializeClass(OutlineManagerMixin)
+InitializeClass(OutlineManager)
 
-class SubtopicsPropertyMixin:
-    """
-    I determine when to display subtopics on a page.
-    """
-    security = ClassSecurityInfo()
 
-    def subtopicsEnabled(self,**kw):
-        """
-        Decide in a complicated way if this page should display it's subtopics.
-
-        First, if the folder has a show_subtopics property (can acquire)
-        and it's false, we will never show subtopics.  Otherwise look for
-        a show_subtopics property
-        - in REQUEST
-        - on the current page
-        - on our primary ancestor pages, all the way to the top,
-        and return the first one we find. Otherwise return true.
-        """
-        prop = 'show_subtopics'
-        if getattr(self.folder(),prop,1):
-            if kw.has_key(prop):
-                return kw[prop] and 1
-            elif hasattr(self,'REQUEST') and hasattr(self.REQUEST,prop):
-                return getattr(self.REQUEST,prop) and 1
-            elif hasattr(self.aq_base,prop):
-                return getattr(self,prop) and 1
-            elif self.primaryParent():
-                # poor caching
-                try: return self.primaryParent().subtopicsEnabled() 
-                except:
-                    # experimental: support all-brains
-                    try: return self.primaryParent().getObject().subtopicsEnabled() 
-                    except: # XXX still run into errors here, investigate
-                        BLATHER('DEBUG: error in subtopicsEnabled for %s, primaryParent is: %s'\
-                             % (self.id(),`self.primaryParent()`))
-                        return not (getattr(getattr(self,'REQUEST',None),
-                                            'zwiki_displaymode',
-                                            None) == 'minimal')
-            else:
-                #return not (getattr(getattr(self,'REQUEST',None),
-                #                    'zwiki_displaymode',
-                #                    None) == 'minimal')
-                return 1
-        else:
-            return 0
-
-    def subtopicsPropertyStatus(self):
-        """
-        Get the status of the show_subtopics property on this page.
-
-        no property:    -1 ("default")
-        true property:   1 ("always")
-        false property:  0 ("never")
-        """
-        if not hasattr(self.aq_base,'show_subtopics'): return -1
-        else: return self.show_subtopics and 1
-
-    def setSubtopicsPropertyStatus(self,status,REQUEST=None):
-        """
-        Set, clear or remove this page's show_subtopics property.
-
-        Same values as getSubtopicsStatus.
-        """
-        props = map(lambda x:x['id'], self._properties)
-        if status == -1:
-            if 'show_subtopics' in props:
-                self.manage_delProperties(ids=['show_subtopics'],
-                                          REQUEST=REQUEST)
-        elif status:
-            if not 'show_subtopics' in props:
-                self.manage_addProperty('show_subtopics',1,'boolean',
-                                        REQUEST=REQUEST)
-            else:
-                self.manage_changeProperties(show_subtopics=1,
-                                             REQUEST=REQUEST)
-        else:
-            if not 'show_subtopics' in props:
-                self.manage_addProperty('show_subtopics',0,'boolean',
-                                        REQUEST=REQUEST)
-            else:
-                self.manage_changeProperties(show_subtopics=0,
-                                             REQUEST=REQUEST)
-
-InitializeClass(SubtopicsPropertyMixin)
-
-class OutlineRenderingMixin:
+class OutlineRendering:
     """
     I present various parts of the wiki outline as HTML.
 
@@ -579,7 +645,6 @@ class OutlineRenderingMixin:
         here = (here and unquote(here)) or self.pageName()
         return self.contentspage(
             self.renderNesting(combos, here),
-            #self.renderNestingX(combos, here),  # dan's skinnable version
             singletons,
             REQUEST=REQUEST)
 
@@ -630,36 +695,30 @@ class OutlineRenderingMixin:
         contentslink = \
           '<a href="%s" title="show wiki contents">%s contents</a>' \
           % (contentsurl, self.folder().title)
+        #with a contents link
         #return '<small><ul>%s\n%s\n</ul></small>' % (contentslink,hierarchy)
-        #XXX try no contents link in context
-        return '<small>%s\n</small>' % (hierarchy)
+        #XXX temp
+        if self.usingPloneSkin():
+            return '%s\n' % (hierarchy)
+        else:
+            return '<small>%s\n</small>' % (hierarchy)
 
     security.declareProtected(Permissions.View, 'contextX')
     def contextX(self, REQUEST=None, with_siblings=0):
         """
-        Return this page's context as nesting structure in a dictionary.
+        Return this page's context information as a skin-renderable structure.
 
-        Like context, but allows a skin template to control the rendering,
+        Like context, but allows a skin template to control the rendering.
+        See nestingAsRenderList.
         """
-        # get the nesting structure
-        here = self.pageName()
         if with_siblings:
-            #nesting = WikiNesting(self.folder()).get_up_and_back(here)
             nesting = self.ancestorsAndChildrenNesting()
         else:
-            # why does the above require a nesting and not this one ?
-            # nesting = self.get_ancestors()
-            #nesting = WikiNesting(self.folder()).get_ancestors(here,self)
             nesting = self.ancestorsNesting()
-            # XXX looks like cruft
-            #if (len(nesting) == 0  or
-            #    (len(nesting) == 1 and len(nesting[0]) == 1)):
-            #    return {'contentsUrl':self.contentsUrl(),
-            #        'hierarchy':{}}
 
-        # format and link it
         # backwards compatibility: in case of an old editform template
         # which shows context, include the new page name at the bottom (unlinked)
+        here = self.pageName()
         if REQUEST.has_key('page') and REQUEST['page'] is not here:
             here = REQUEST['page']
             nesting = deepappend(nesting, here)
@@ -667,17 +726,15 @@ class OutlineRenderingMixin:
         else:
             suppress_hyperlink=0
 
-        hierarchy = self.renderNestingX( nesting, here )
-            
+        # convert to a render list
+        renderlist = self.nestingAsRenderList(nesting, here)
         # special case: if parent seems to be missing, reset XXX
-        if len(hierarchy) == 2 :
+        if len(renderlist) == 2 :
             self.setParents([])
             self.index_object()
-            hierarchy = self.renderNestingX(
-                nesting, here)
-                
-        # if a SiteMap page exists, point the contents link there
-        return {'contentsUrl':self.contentsUrl(), 'hierarchy':hierarchy}
+            renderlist = self.nestingAsRenderList(nesting, here)
+
+        return {'contentsUrl':self.contentsUrl(), 'hierarchy':renderlist}
 
     security.declareProtected(Permissions.View, 'children')
     def children(self):
@@ -691,32 +748,38 @@ class OutlineRenderingMixin:
             return ''
 
     security.declareProtected(Permissions.View, 'offspring')
-    def offspring(self, REQUEST=None, info=None, exclude_self=0):
+    def offspring(self, REQUEST=None, info=None, exclude_self=0, depth=None):
         """
         Return HTML displaying all my offspring.
         """
         here = self.pageName()
         return self.renderNesting(
-            self.offspringNesting(),
-            here,
-            suppress_current=exclude_self)
+            self.offspringNesting(depth=depth),here,suppress_current=exclude_self)
 
     security.declareProtected(Permissions.View, 'subtopics')
-    def subtopics(self,deep=0):
+    def subtopics(self, REQUEST=None, **kw):
         """
-        Return HTML showing my subtopics (offspring with different formatting).
+        Render my subtopics as a HTML fragment, using a template.
+
+        Allows the template to be selected by subtopics_style folder
+        property. If there is no property or the specified template
+        does not exist, uses the built-in subtopics_outline.
+        Some overlap with the show_subtopics property.
+
+        Pass any arguments, like max depth, through to the template.
         """
-        subtopics = ((deep and self.offspring(exclude_self=1)) or
-                     self.children())
-        if subtopics:
-            return '\n\n<a name="subtopics"><br /></a>\n<p><table id="subtopicspanel"><tr><td><b>subtopics:</b>\n%s</td></tr></table>' \
-                   % (subtopics) #XXX i18n
-        else: return ''
+        DEFAULTSTYLE = 'outline'
+        style = getattr(self,'subtopics_style',None)
+        if not (style and self.hasSkinTemplate('subtopics_'+style)):
+            style = DEFAULTSTYLE
+        return self.getSkinTemplate('subtopics_'+style)(self,REQUEST,**kw)
 
     security.declareProtected(Permissions.View, 'navlinks')
     def navlinks(self):
         """
         Return HTML for my next/previous/up links.
+
+        XXX not used ?
         """
         none = 'none'
         t = ''
@@ -753,31 +816,73 @@ class OutlineRenderingMixin:
           (backwards compatibility for old editforms)
         - if suppress_current is true, here will not be shown at all
         - did, got & indent are for recursion, callers should not use
-
-        """
-        #XXX cleanup
         
+        Do we need all this complicated code, can't we do it in the skin
+        template ? I think so, except for two issues: for very large
+        nestings the python version might be perceptibly quicker; and,
+        it's easier to recurse with python. See also nestingAsRenderList. 
+        """
+        # XXX oh yeah.. kludgorama
+        def renderContentsLink(page):
+            """Render a link to page, suitable for contents or context."""
+            # quicker than renderLinkToPage, since we know it exists:
+            wikiurl = self.wiki_url()
+            def quicklink(page):
+                return '<a href="%s/%s" name="%s">%s</a>' \
+                       % (wikiurl,self.canonicalIdFrom(page),quote(page),
+                          self.formatWikiname(page))
+            if here and page == here: 
+                if enlarge_current:
+                    # just assume we are in the page header, and link to
+                    # backlinks as well as enlarging
+                    #return '<big><big><big><big><strong>%s</strong></big></big></big></big>' % \
+                    return '<h1 style="display:inline;">%s</h1>' % \
+                           ('<a href="%s/%s/backlinks" title="%s" name="%s">%s</a>' % \
+                            (wikiurl,
+                             self.canonicalIdFrom(page),
+                             _("which pages link to this one ?"),
+                             page,
+                             self.formatWikiname(page)))
+                else:
+                    # just highlight "here"
+                    return '%s <b><-- %s.</b>' % \
+                           ((suppress_hyperlink and page) # another special case
+                            or quicklink(page),
+                            _("You are here"))
+            else:
+                #return self.renderLinkToPage(page,name=page)
+                return quicklink(page)
+            
+        #XXX cleanup
         if suppress_current and nesting[0] == here: # a single childless page
             return ''
         if did is None: did = []
         if got is None:
-            got = ['<ul class="aqtree3clickable">'] #SKWM 
+            # XXX trying hard to turn off bullets in plone!
+            got = ['<ul style="list-style:none; margin-left:0;" class="aqtree3clickable">']
             recursing = 0
         else:
             recursing = 1
         for n in nesting:
             if type(n) == ListType:
                 if not (n[0]==here and suppress_current): #XXX temp
-                    got.append('%s <li>[%s]' % (indent,n[0]))
+                    got.append('%s <li style="list-style:none" class="aqtree3clickable">%s' % (indent,renderContentsLink(n[0])))
                 if len(n) > 1:
                     if not (n[0]==here and suppress_current): #XXX temp
                         got.append('<ul>')
                     for i in n[1:]:
                         if type(i) == ListType:
                             got = self.renderNesting(
-                                [i],here,did=did,got=got,indent=indent+' ')
+                                [i],
+                                here=here,
+                                enlarge_current=enlarge_current,
+                                suppress_hyperlink=suppress_hyperlink,
+                                suppress_current=suppress_current,
+                                did=did,
+                                got=got,
+                                indent=indent+' ')
                         else:
-                            got.append('%s <li>[%s]</li>' % (indent,i))
+                            got.append('%s <li style="list-style:none" class="aqtree3clickable">%s</li>' % (indent,renderContentsLink(i)))
                     if not (n[0]==here and suppress_current): #XXX temp
                         got.append("</ul>")
                 else:
@@ -785,49 +890,79 @@ class OutlineRenderingMixin:
                 if not (n[0]==here and suppress_current):
                     got.append('%s </li>' % indent)
             else:
-                got.append('%s <li>[%s]</li>' % (indent,n))
-        if recursing: return got
+                got.append('%s <li style="list-style:none" class="aqtree3clickable">%s</li>' % (indent,renderContentsLink(n)))
 
-        # finish up, do pretty printing options and wiki links
-        got.append("</ul>")
-        t = join(got, "\n")
-        if here:
-            if enlarge_current:
-                t = re.sub(r'(\[%s\])' % re.escape(here),
-                           r'<big><big><big><big><strong>\1</strong></big></big></big></big>',
-                           t)
-                # XXX temporary kludge.. assume we are in the page header here
-                t = re.sub(r'(\[%s\])' % re.escape(self.pageName()),
-                           '<a href="%s/backlinks" title="%s">%s</a>' % (
-                               self.page_url(),
-                               _("which pages link to this one ?"),
-                               self.formatWikiname(self.pageName())),
-                           t)
-            elif SHOW_CURRENT_PAGE_IN_CONTENTS:
-                t = re.sub(r'(\[%s\])' % re.escape(here),
-                           r'\1 <b><-- %s.</b>' % _("You are here"),
-                           t)
-            if suppress_hyperlink:
-                t = re.sub(r'(\[%s\])' % re.escape(here), r'!\1', t)
-        #t = self.renderLinksIn(t) # too expensive for now.. do it on the cheap
-        wikiurl = self.wiki_url()
-        def quicklink(match):
-            page = match.group(1)
-            return '<a href="%s/%s" name="%s">%s</a>' \
-                   % (wikiurl,self.canonicalIdFrom(page),quote(page),
-                      self.formatWikiname(page))
-        t = re.sub(bracketedexpr,quicklink,t)
-        return t
+        if recursing:
+            return got
+        else:
+            got.append("</ul>")
+            return join(got, "\n")
 
-    security.declareProtected(Permissions.View, 'renderNestingX')
-    def renderNestingX(self, nesting, here=None, suppress_current=0,
+    security.declareProtected(Permissions.View, 'nestingAsRenderList')
+    def nestingAsRenderList(self, nesting, here=None, suppress_current=0,
                       did=None, got=None, indent=''):
         """
-        Unpack a nesting structure into a list.
+        Convert a nesting structure to a skin-renderable intermediate form.
 
-        Like renderNesting, but supports contextX which allows a skin
-        template to do the rendering.
+        This was renderNestingX in Dan McMullen's skin refactoring
+        (http://zwiki.org/WikipageX). The intermediate form is a flat list
+        which a skin template can render without needing recursion.
+
+        Example: a chunk of page hierarchy like::
+
+          AnnoyingQuote
+            AnnoyingQuoteArchive
+              AngryDenial
+            AnnoyingQuoteDiscussion
+
+        which is stored as a nesting like::
+
+          [['AnnoyingQuote', ['AnnoyingQuoteArchive', 'AngryDenial'], 'AnnoyingQuoteDiscussion']]        
+
+        gets converted to a render list like::
+
+          [
+          {'type': '+'}, 
+          {'href': 'zwikib/AnnoyingQuote', 'type': '=', 'page': 'AnnoyingQuote', 'name': 'AnnoyingQuote'}, 
+          {'type': '+'}, 
+          {'href': 'zwikib/AnnoyingQuoteArchive', 'type': '=', 'page': 'AnnoyingQuoteArchive', 'name': 'AnnoyingQuoteArchive'}, 
+          {'type': '+'}, 
+          {'href': 'zwikib/AngryDenial', 'type': '=', 'page': 'AngryDenial', 'name': 'AngryDenial'}, 
+          {'type': '-'}, 
+          {'href': 'zwikib/AnnoyingQuoteDiscussion', 'type': '=', 'page': 'AnnoyingQuoteDiscussion', 'name': 'AnnoyingQuoteDiscussion'}, 
+          {'type': '-'}, 
+          {'type': '-'}
+          ]
+
+        which can be rendered by some TAL like::
+
+          <span tal:define="ctx python: here.contextX(request)">
+            <a href="contentsUrl" title="show wiki contents" accesskey="c"
+              tal:attributes="href ctx/contentsUrl"
+              tal:content="python: container.title + ' contents'">contents</a>
+            <span tal:repeat="ii ctx/hierarchy" tal:omit-tag="">
+              <span tal:condition="python: '+' in ii['type']"
+                tal:replace="structure string:<ul style=&quot;margin: 0;&quot;>">ul</span>
+              <span tal:condition="python: '=' in ii['type']">
+                <li tal:condition="python: '!' in ii['type']">
+                  <big><big><big><big><strong><a href="href" name="name"
+                  tal:attributes="href python: ii['href'] + '/backlinks'; name ii/name"
+                  tal:content="ii/page">page</a></strong></big></big></big></big>
+                  <span tal:condition="python: '.' in ii['type']"> ...</span>
+                </li>
+                <li tal:condition="python: '!' not in ii['type']">
+                  <a href="href" name="name"
+                    tal:attributes="href ii/href; name ii/name"
+                    tal:content="ii/page">page</a>
+                    <span tal:condition="python: '.' in ii['type']"> ...</span>
+                </li>
+              </span>
+              <span tal:condition="python: '-' in ii['type']"
+                tal:replace="structure string:</ul>">ultag</span>
+            </span>
+          </span>
         """
+
         if suppress_current and nesting[0] == here: # a single childless page
             return []
         if did is None: did = []
@@ -846,7 +981,7 @@ class OutlineRenderingMixin:
                         got.append( {'type':'+'} )
                     for i in n[1:]:
                         if type(i) == ListType:
-                            got = self.renderNestingX(
+                            got = self.nestingAsRenderList(
                                 [i],here,did=did,got=got,indent=indent+' ')
                         else:
                             t = (i==here and '=!' or '=')
@@ -875,13 +1010,13 @@ class OutlineRenderingMixin:
     # backwards compatibility
     map = contents
 
-InitializeClass(OutlineRenderingMixin)
+InitializeClass(OutlineRendering)
 
-class OutlineSupport(
-    OutlineManagerMixin, 
-    ParentsPropertyMixin,
-    SubtopicsPropertyMixin,
-    OutlineRenderingMixin
+class PageOutlineSupport(
+    ParentsProperty,
+    ShowSubtopicsProperty,
+    OutlineManager, 
+    OutlineRendering
     ):
     """
     I make a page aware of it's place within the overall wiki outline
