@@ -25,21 +25,85 @@ except:
     ZOPEVERSION = (9,9,9) # (cvs)
 
 from Products.ZWiki import __version__
-from Defaults import PREFER_USERNAME_COOKIE, PAGE_METADATA, \
-     SHOW_CURRENT_PAGE_IN_CONTENTS
+from Defaults import PREFER_USERNAME_COOKIE, PAGE_METADATA, BORING_PAGES
 import Permissions
-from I18nSupport import _
+from I18n import _
+
+# adapt to zope 2.9 transaction api
+try: from transaction import get as get_transaction
+except ImportError: get_transaction = get_transaction
 
 
-class Utils:
+class PageUtils:
     """
-    Miscellaneous utilities mixin for ZWikiPage.
+    Miscellaneous utility methods for zwiki pages.
     """
     security = ClassSecurityInfo()
     security.declareObjectProtected('View')
+
+    ######################################################################
+    # misc security utilities
+    
     def checkPermission(self, permission, object):
         return getSecurityManager().checkPermission(permission,object)
 
+    # Zwiki implements a little extra access control in addition to Zope's
+    # permissions, to allow more lightweight security based on cookies etc.
+    def checkSufficientId(self, REQUEST=None):
+        REQUEST = REQUEST or getattr(self,'REQUEST',None)
+        return (self.requestHasUsername(REQUEST) or
+                not getattr(self,'edits_need_username',0))
+
+    # used this in standard templates for while, so must keep it for
+    # backwards compatibility. The name is no longer accurate.
+    userIsIdentified = checkSufficientId
+
+    def requestHasUsername(self,REQUEST=None):
+        """
+        Check REQUEST has either an authenticated user or a username cookie.
+        """
+        REQUEST = REQUEST or getattr(self,'REQUEST',None)
+        username = self.usernameFrom(REQUEST)
+        return (username and username != REQUEST.REMOTE_ADDR)
+
+    def usernameFrom(self, REQUEST=None, ip_address=1):
+        """
+        Get the best available user id from the current visitor's REQUEST.
+
+        We use the first of:
+        - a MAILIN_USERNAME set by mailin.py
+        - an authenticated username
+        - a zwiki_username cookie
+        - the client's IP address, unless disabled
+
+        Or, if PREFER_USERNAME_COOKIE in Defaults.py is true, we'll let
+        the cookie take precedence. This means an authenticated user could
+        pretend to be someone else by setting the cookie, but it's useful
+        for this case: a simple community wiki protected by a single
+        common login, multiple users distinguished by their
+        zwiki_username.
+
+        We'll find REQUEST ourself if necessary (helps backwards compatibility
+        with old skin templates).
+        """
+        REQUEST = REQUEST or getattr(self,'REQUEST',None)
+        if not REQUEST: return ''
+        authenticated_name = None
+        user = REQUEST.get('AUTHENTICATED_USER')
+        if user:
+            authenticated_name = user.getUserName()
+            if authenticated_name == str(user.acl_users._nobody):
+                authenticated_name = None
+        cookie_name = REQUEST.cookies.get('zwiki_username',None)
+        mailin_name = REQUEST.get('MAILIN_USERNAME',None)
+        ip_addr = ip_address and REQUEST.REMOTE_ADDR or ''
+        if PREFER_USERNAME_COOKIE:
+            return mailin_name or cookie_name or authenticated_name or ip_addr
+        else:
+            return mailin_name or authenticated_name or cookie_name or ip_addr
+
+    ######################################################################
+
     def size(self):
         """
         Give the size of this page's text.
@@ -62,43 +126,57 @@ class Utils:
 
     def summary(self,size=200,paragraphs=1):
         """
-        Give a short summary of this page's content.
+        Give a short plaintext summary of this page's content.
 
-        Take the first N paragraphs of the document part; cut to size
-        characters & replace the last word with ellipsis if necessary.
+        We generate this by excerpting the first paragraph.
+        Specifically, we take the page's document part, strip html tags,
+        and return up to the specified number of characters or paragraphs,
+        replacing the last word with an ellipsis if we had to truncate.
         """
         size, paragraphs = int(size), int(paragraphs)
         t = self.documentPart()
+        t = re.sub(r'<(?=\S)[^>]+>','',t).strip() # strip html tags
         if paragraphs: t = join(split(t,'\n\n')[:paragraphs],'\n\n')
         if len(t) > size:
             t = t[:size]
             t = re.sub(r'\w*$',r'',t) + '...'
         return html_quote(t)
 
+    def renderedSummary(self,size=500,paragraphs=1):
+        """
+        Give a summary of this page's content, as rendered html.
+
+        Similar to summary(), but this one tries to apply the page's
+        formatting rules and do wiki linking. We remove any enclosing <p>.
+        """
+        return re.sub(r'(?si)^<p>(.*)</p>\n?$', r'\1',
+            self.renderLinksIn(
+            self.pageType().format(
+            self.summary(size=size, paragraphs=paragraphs))))
+
     security.declareProtected(Permissions.View, 'excerptAt')
-    def excerptAt(self, expr, size=100, highlight=1):
+    def excerptAt(self, expr, size=100, highlight=1, text=None):
         """
-        Return an excerpt of this page at the first occurence of expr.
+        Return a highlighted search result excerpt from this page (or text).
 
-        Useful when presenting search results.  Does a straight
-        case-insensitive string search, after first removing any *
-        wildcard character (this may not agree exactly with the catalog
-        index's search). SGML tags are quoted.
+        This method searches this page's text, or the provided text, for
+        the first occurrence of expr (cleaned up) and returns the
+        surrounding text chunk, html quoted, and optionally with the
+        matches enclosed in styled spans. If no match is found, it just
+        returns a chunk from the beginning.
 
-        If no match is found or expr is blank, returns an empty string.
-
-        When highlight is true, each match within the excerpt is enclosed
-        in a span with class "hit". This is not 100% reliable because of
-        html quoting and because we don't handle wildcards.
-        
+        It should mimic the search strategy of SearchPage, but that
+        depends on catalog configuration and it currently doesn't do a
+        very good job, so the excerpts and highlights can be misleading.
         """
+        text = text or self.text()
         string = re.sub(r'\*','',expr)
-        m = re.search(r'(?i)'+re.escape(string),self.text())
+        m = re.search(r'(?i)'+re.escape(string),text)
         if m and string:
             middle = (m.start()+m.end())/2
             exstart = max(middle-size/2-1,0)
-            exend = min(middle+size/2+1,len(self.text()))
-            excerpt = html_quote(self.text()[exstart:exend])
+            exend = min(middle+size/2+1,len(text))
+            excerpt = html_quote(text[exstart:exend])
             if highlight:
                 excerpt = re.sub(
                     r'(?i)'+re.escape(html_quote(string)),
@@ -108,7 +186,7 @@ class Utils:
                     excerpt)
             return excerpt
         else:
-            return ''
+            return html_quote(text[:size])
 
     def metadataFor(self,page):
         """
@@ -122,47 +200,22 @@ class Utils:
         Warning: fields such as the parents list may be
         copies-by-reference, and should not be mutated.
         """
-        class PageBrain(SimpleItem):
+        class PageBrain(SimpleItem): # XXX why a SimpleItem ?
             def __init__(self,obj): self._obj = obj
             def getObject(self): return self._obj
         brain = PageBrain(page)
         for attr in PAGE_METADATA+['page_url','title_or_id']:
-            # don't acquire these properties
-            # XXX messy.. aq_base won't be present in tests
-            # XXX temp protection for unpickler import error when page
-            # type class is missing
+            # protect against import or attr errors when a page type has
+            # been uninstalled
             try:
+                # don't acquire these properties
+                # aq_base won't be present during unit tests.. messy
                 setattr(brain, attr,
                         absattr(getattr(getattr(page,'aq_base',page),attr,None)))
-                #setattr(brain,attr+'__roles__',None) # not needed ?
-            except ImportError:
+            except (ImportError, AttributeError):
                 setattr(brain, attr, None)
-                # didn't work
-                #Traceback (most recent call last):
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/ZPublisher/Publish.py", line 101, in publish
-                #        request, bind=1)
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/ZPublisher/mapply.py", line 88, in mapply
-                #        if debug is not None: return debug(object,args,context)
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/ZPublisher/Publish.py", line 39, in call_object
-                #        result=apply(object,args) # Type s<cr> to step into published object.
-                #    File "/zope1/Products/ZWiki/UI.py", line 415, in recentchanges
-                #        return form(self,REQUEST,body=dtmlpart(self,REQUEST))
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/App/special_dtml.py", line 62, in __call__
-                #        (self,)+args[1:],kw)
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/DocumentTemplate/DT_String.py", line 474, in __call__
-                #        try: result = render_blocks(self._v_blocks, md)
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/DocumentTemplate/DT_Let.py", line 76, in render
-                #        return render_blocks(self.section, md)
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/DocumentTemplate/DT_In.py", line 703, in renderwob
-                #        try: append(render(section, md))
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/DocumentTemplate/DT_Try.py", line 140, in render
-                #        return self.render_try_except(md)
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/DocumentTemplate/DT_Try.py", line 174, in render_try_except
-                #        return render_blocks(handler, md)
-                #    File "/usr/local/src/Zope-2.7.1-b2/lib/python/DocumentTemplate/DT_Var.py", line 280, in render
-                #        val = fmt % val
-                #TypeError: not all arguments converted during string formatting
-        brain.linkTitle = '' #XXX not using now.. leave blank so tests pass
+        #XXX not using now.. leave blank so tests pass
+        brain.linkTitle = '' 
         return brain
 
     def ensureCompleteMetadataIn(self,brain):
@@ -215,56 +268,6 @@ class Utils:
     def urlunquote(self, text):
         return unquote(text)
 
-    security.declareProtected(Permissions.View, 'usernameFrom')
-    def usernameFrom(self, REQUEST=None, ip_address=1):
-        """
-        Get the best available user id from the current visitor's REQUEST.
-
-        We use the first of:
-        - a MAILIN_USERNAME set by mailin.py
-        #- an authenticated CMF user's fullname property #XXX check old code
-        - an authenticated username
-        - a zwiki_username cookie
-        - the client's IP address, unless disabled
-
-        Or, if PREFER_USERNAME_COOKIE in Defaults.py is true, we'll let
-        the cookie take precedence. This means an authenticated user could
-        pretend to be someone else by setting the cookie, but it's useful
-        for this case: a simple community wiki protected by a single
-        common login, multiple users distinguished by their
-        zwiki_username.
-
-        We'll find REQUEST ourself if necessary (helps backwards compatibility
-        with old skin templates).
-        """
-        REQUEST = REQUEST or getattr(self,'REQUEST',None)
-        if not REQUEST: return ''
-        authenticated_name = None
-        user = REQUEST.get('AUTHENTICATED_USER')
-        if user:
-            #if self.inCMF():
-            #  authenticated_name = self.folder().portal_membership.getMemberById(user.getUserName()).fullname or ''
-            #else:
-            authenticated_name = user.getUserName()
-            
-            if authenticated_name == str(user.acl_users._nobody):
-                authenticated_name = None
-        cookie_name = REQUEST.cookies.get('zwiki_username',None)
-        mailin_name = REQUEST.get('MAILIN_USERNAME',None)
-        ip_addr = ip_address and REQUEST.REMOTE_ADDR or ''
-        if PREFER_USERNAME_COOKIE:
-            return mailin_name or cookie_name or authenticated_name or ip_addr
-        else:
-            return mailin_name or authenticated_name or cookie_name or ip_addr
-
-    security.declareProtected(Permissions.View, 'requestHasSomeId')
-    def requestHasSomeId(self,REQUEST):
-        """
-        Check REQUEST has either a non-anonymous user or a username cookie.
-        """
-        username = self.usernameFrom(REQUEST)
-        return (username and username != REQUEST.REMOTE_ADDR)
-
     def __repr__(self):
         return ("<%s %s at 0x%s>"
                 % (self.__class__.__name__, `self.id()`, hex(id(self))[2:]))
@@ -321,31 +324,17 @@ class Utils:
         return self.urlForPageOrDefault('FrontPage',self.wikiUrl())
 
     security.declareProtected(Permissions.View, 'contentsUrl')
-    def contentsUrl(self):
+    def contentsUrl(self, scroll=1):
         """
-        Return the url of the SiteMap page or wiki contents method.
-        
-        We used to put the current page name as a target and scroll there,
-        but in a large public wiki that generates a lot of unique
-        expensive urls for robots. I'm assuming robots are too dumb
-        not to crawl both url#target1 and url#target2.
-
-        Perhaps the contents method could check HTTP_REFERER and do a
-        redirect, only if you are a human with a cookie ? brrr, confusing..
-        There's something to be said for not having the contents scroll
-        out from under you anyhow. Try without it for a bit.
+        Return the url of zwiki's contents method.
         
+        By default we put the current page name as a target and scroll
+        there, note in a large public wiki that generates a lot of unique
+        urls for robots, if robots are dumb enough to crawl both
+        url#target1 and url#target2.
         """
-        if SHOW_CURRENT_PAGE_IN_CONTENTS:
-            quotedname = quote(self.pageName())
-            if self.pageWithId('SiteMap'): 
-                return '%s/SiteMap?here=%s#%s' \
-                % (self.wiki_url(),quotedname,quotedname)
-            else: 
-                return '%s/contents#%s' % (self.page_url(),quotedname)
-        else:
-            if self.pageWithId('SiteMap'): return self.wikiUrl() + '/SiteMap'
-            else: return self.defaultPageUrl() + '/contents'
+        return self.defaultPageUrl() + '/contents' + \
+               (scroll and ('#%s' % quote(self.pageName())) or '')
 
     security.declareProtected(Permissions.View, 'changesUrl')
     def changesUrl(self):
@@ -356,18 +345,6 @@ class Utils:
         p = self.pageWithName('UserDiscussion') or self.pageWithName('GeneralDiscussion')
         return (p and p.pageUrl()) or ''
 
-    security.declareProtected(Permissions.View, 'trackerUrl')
-    def trackerUrl(self):
-        return self.urlForDtmlPageOrMethod('IssueTracker','issuetracker')
-
-    security.declareProtected(Permissions.View, 'filterUrl')
-    def filterUrl(self):
-        return self.urlForDtmlPageOrMethod('FilterIssues','filterissues')
-
-    security.declareProtected(Permissions.View, 'filterUrl')
-    def issueBrowserUrl(self):
-        return self.urlForDtmlPageOrMethod('IssueBrowser','issuebrowser')
-
     security.declareProtected(Permissions.View, 'indexUrl')
     def indexUrl(self):
         return self.urlForDtmlPageOrMethod('AllPages','') #'allpages')
@@ -447,6 +424,14 @@ class Utils:
         return int(self.getPhysicalRoot().ZopeTime() -
                    self.lastEditTime())
 
+    security.declareProtected(Permissions.View, 'lastEditInterval')
+    def lastEditIntervalInHours(self):
+        """
+        return the number of hours since last edit
+        """
+        return int((self.getPhysicalRoot().ZopeTime() -
+                    self.lastEditTime()) * 24)
+
     security.declareProtected(Permissions.View, 'asAgeString')
     def asAgeString(self,time):
         """
@@ -501,11 +486,31 @@ class Utils:
         Renders without the skin, passes REQUEST in case authentication is
         needed, fails silently if page does not exist.
         """
+        REQUEST = REQUEST or self.REQUEST
         p = self.pageWithNameOrId(page)
         if p: return p(bare=1,REQUEST=REQUEST, **kw)
         else: return ''
         
-InitializeClass(Utils)
+    def isBoring(self):
+        """
+        Is this page one which should be quieter, eg test pages ?
+
+        Boring pages are pages which we don't usually want to see in blog
+        listings, rss feeds etc. (?) or hear mail from unless subscribed
+        directly. These are TestPage, SandBox and their offspring, by
+        default. You can configure different pages in a boring_pages lines
+        folder property, one per line.
+        """
+        boring = getattr(self,'boring_pages', BORING_PAGES)
+        if self.pageName() in boring:
+            return 1
+        ancestors = self.ancestorsAsList()
+        for p in boring:
+            if p in ancestors:
+                return 1
+        return 0
+
+InitializeClass(PageUtils)
 
 
 # generic utilities
@@ -519,6 +524,11 @@ def WARN(*args):
     for arg in args: tmp.append(str(arg))
     zLOG.LOG('ZWiki',zLOG.WARNING,' '.join(tmp))
 
+def DEBUG(*args):
+    tmp = []
+    for arg in args: tmp.append(str(arg))
+    zLOG.LOG('ZWiki',zLOG.DEBUG,' '.join(tmp))
+
 def formattedTraceback():
     type,val,tb = sys.exc_info()
     try:     return join(traceback.format_exception(type,val,tb),'')
@@ -730,7 +740,7 @@ def isEmailAddress(s):
     """
     True if s looks like an email address.
     """
-    if s and '@' in s: return 1
+    if type(s) is StringType and '@' in s: return 1
     else: return 0
 
 def isUsername(s):
@@ -739,3 +749,57 @@ def isUsername(s):
     """
     return not isEmailAddress(s)
 
+# utilities for managing lists of callable actions, "hooks" as they're
+# called in emacs.  Certain core zwiki methods allow themselves to be
+# extended at run-time by "hooking in" additional methods/fns, usually by
+# plugins.  Cf aspect-oriented programming, method annotations, ...
+#
+# Declare a hook list before a method like this:
+#
+#   # allow extra actions to be added to this method
+#   global methodname_hooks
+#   methodname_hooks = []
+#
+# and have the method call the hooks like this:
+#
+#   callHooks(methodname_hooks, self)
+#
+# then other code can add a hook like this:
+#
+#   from Products.ZWiki.Utils import addHook
+#   from Products.ZWiki.Modulename import methodname_hooks
+#   addHook(methodname_hooks, hookmethod)
+#
+# the hook needs to accept the type of argument callHooks gives it.
+# see eg Admin.py:upgrade and plugins/tracker/__init__.py
+
+#def declareHook(method):
+#    """
+#    Declare a hook list for method, to allow extra actions to be registered.
+#    """
+#    import exec
+#    exec (method+'_hooks = []')
+#    exec ('global '+method+'_hooks')
+
+def addHook(hooks, fn):
+    """
+    Add a function to a list of hook actions to be called.
+    """
+    hooks.append(fn)
+
+def callHooks(hooks, arg):
+    """
+    Call each of a list of functions with arg, returning any error code.
+
+    Hook functions are called with one argument.  We catch and log any
+    exceptions, and return the last non-null return code if any.
+    """
+    err = None
+    for hook in hooks:
+        try:
+            err = hook(arg) or err
+        except:
+            BLATHER(
+                'could not call hook, skipping (traceback follows)\n%s' % (
+                formattedTraceback()))
+    return err
